
package use_trafficintersection;

import java.util.ArrayList;
import java.util.Random;
import use_trafficintersection.algorithm.Car;
import use_trafficintersection.algorithm.Main;
import use_trafficintersection.algorithm.Lights;

/**
 *
 * @author Sjoerd Westendorp
 */
public class AutoTester {
    public static boolean useAutoTester = false;
    public static boolean unitTest = false;
    public static boolean autoTest = false;
    public String unitTestName = "";
    
    public static long TEST_CARCOUNT = 2*Main.TIME_SPENT_LIST_MAX_SIZE;
    
    // Other objects
    Main main;
    Renderer renderer;
    Lights lights;
    
    // Algo Tests
    public int algo_num = -1;
    public String algo = "";
    public static final String ALGOS_NORMAL[] = new String[]{
	"cost_auto_pl",
	"vac"
    };
    public static final String ALGOS_DEBUG[] = new String[]{
	"vac",
	"cost_auto_pl",
	"cost_auto",
	"cost",
	"fixed"
    };
    public static final String ALGOS[] = ALGOS_NORMAL;
    public long seed = 0;
    
    // Output
    public static ArrayList<String> output = new ArrayList<>();
    
    // -------------------------------
    
    public AutoTester(Main m, Renderer r,Lights l){
	main = m;
	renderer = r;
	lights = l;
    }
    
    public void reset(){
	useAutoTester = false;
	unitTest = false;
	autoTest = false;
	unitTestName = "";
	algo_num = -1;
	algo = Main.STANDARD_ALGO;
	autoTestReset();
    }
    
    // -------------------------------
    
    public void hotkey(int num){
	if(num == 0){
	    reset(); main.reset();
	    renderer.renderWorld.sliderBiasU.state = renderer.renderWorld.sliderBiasU.start;
	    renderer.renderWorld.sliderBiasD.state = renderer.renderWorld.sliderBiasD.start;
	    renderer.renderWorld.sliderBiasL.state = renderer.renderWorld.sliderBiasL.start;
	    renderer.renderWorld.sliderBiasR.state = renderer.renderWorld.sliderBiasR.start;
	    renderer.renderWorld.sliderCars.state = renderer.renderWorld.sliderCars.start;
	    renderer.renderWorld.sliderAutos.state = renderer.renderWorld.sliderAutos.start;
	    renderer.renderWorld.sliderPeds.state = renderer.renderWorld.sliderPeds.start;
	    renderer.renderWorld.sliderPacketLoss.state = renderer.renderWorld.sliderPacketLoss.start;
	}
	else if(num == 1){
	    reset(); main.reset();
	    useAutoTester = true; unitTest = true;
	    algo = "cost_auto_pl";
	    unitTestName = algo.toUpperCase();
	}
	else if(num == 2){
	    reset(); main.reset();
	    useAutoTester = true; unitTest = true;
	    algo = "vac";
	    unitTestName = algo.toUpperCase();
	}
	else if(num == 3){
	    reset(); main.reset();
	    useAutoTester = true; unitTest = true;
	    algo = "cost_auto";
	    unitTestName = algo.toUpperCase();
	}
	else if(num == 4){
	    reset(); main.reset();
	    useAutoTester = true; unitTest = true;
	    // Set traffic from one direction
	    renderer.renderWorld.sliderBiasU.state = renderer.renderWorld.sliderBiasU.min;
	    renderer.renderWorld.sliderBiasD.state = renderer.renderWorld.sliderBiasD.min;
	    renderer.renderWorld.sliderBiasL.state = renderer.renderWorld.sliderBiasL.max;
	    renderer.renderWorld.sliderBiasR.state = renderer.renderWorld.sliderBiasR.min;
	    // Set traffic count
	    renderer.renderWorld.sliderCars.state = renderer.renderWorld.sliderCars.start;
	    renderer.renderWorld.sliderAutos.state = renderer.renderWorld.sliderCars.max;
	    renderer.renderWorld.sliderPeds.state = renderer.renderWorld.sliderPeds.start;
	    // Other settings
	    renderer.renderWorld.sliderSimSpeed.state = renderer.renderWorld.sliderSimSpeed.min;
	    // Set algo
	    algo = "cost_auto";
	    unitTestName = "Green Wave";
	}
	else if(num == 5){
	    autoTestStart();
	}
    }
    
    public void test(){
	if(unitTest)
	    testUnit();
	else if(autoTest)
	    testAutoTest();
	else
	    testAlgo();
    }
    
    public void testThis(String algorithm){
	algo = algorithm;
	testAlgoStep();
    }
    
    public void testAlgo(){
	if( algo_num < 0){
	    // Random seed selector
	    Random rand = new Random();
	    seed = rand.nextLong();
	    // No algo set, lets test the first one
	    String str = "Cars: "+((int)renderer.renderWorld.sliderCars.state)+"%";
	    str += " | Autos: "+((int)renderer.renderWorld.sliderAutos.state)+"%";
	    str += " | Loss: "+((int)renderer.renderWorld.sliderPacketLoss.state)+"%";
	    str += " | Peds: "+((int)renderer.renderWorld.sliderPeds.state)+"%";
	    log(str);
	    str = "Bias";
	    str += " | U: "+((int)renderer.renderWorld.sliderBiasU.state)+"%";
	    str += " | L: "+((int)renderer.renderWorld.sliderBiasL.state)+"%";
	    str += " | D: "+((int)renderer.renderWorld.sliderBiasD.state)+"%";
	    str += " | R: "+((int)renderer.renderWorld.sliderBiasR.state)+"%";
	    log(str);
	    str = "Seed: " + seed;
	    str += " | Platoon: "+Car.usePlatooning;
	    log(str);
	    algo_num = 0;
	    testSetupAlgo();
	    
	}
	if( algo_num >= ALGOS.length){
	    useAutoTester = false;
	    log("DONE"); log("");
	    algo_num = -1;
	    renderer.renderWorld.sliderSimSpeed.state = renderer.renderWorld.sliderSimSpeed.min;
	    lights.reset();
	    return; // We are done
	}
	
	// Check wether we are done with this algo
	if(main.carCount >= TEST_CARCOUNT){
	    String s = "["+algo.toUpperCase()+"] ";
	    log(s);
	    s = "    Delay: "+UnitConv.loopsToDispString(main.timeSpent);
	    s += " | Ped: "+UnitConv.loopsToDispString(main.timeSpentPed);
	    log(s);
	    s = "    Accel: "+UnitConv.accelToDispString(main.accel);
	    s += " | PosAccel: "+UnitConv.accelToDispString(main.posaccel);
	    log(s);
	    
	    // Start next one
	    algo_num++;
	    testSetupAlgo();
	}
	
	// Try algorithm
	testAlgoStep();
    }
    
    public void testUnit(){
	testAlgoStep();
    }
    
    private void testAlgoStep(){
	if( algo.equals("fixed") )
	    lights.algoFixed();
	else if( algo.equals("cost") )
	    lights.algoCost(false,false);
	else if( algo.equals("cost_auto") )
	    lights.algoCost(true,false);
	else if( algo.equals("cost_auto_pl") )
	    lights.algoCost(true,true);
	else if( algo.equals("vac") )
	    lights.algoVac();
	else{
	    System.out.print("[ERROR] lights.testAlgoStep -> Incorrect Algo num: "+algo);
	    System.out.println(", max:"+(ALGOS.length)+".");
	}
    }
    
    private void testSetupAlgo(){
	if( algo_num >= ALGOS.length){
	    return; // We are done
	}
	algo = ALGOS[algo_num];
	main.reset();
	renderer.renderWorld.sliderSimSpeed.state = renderer.renderWorld.sliderSimSpeed.max;
	// Random seed selector
	main.rand_uniform.setSeed(seed);
	main.rand_packetloss.setSeed(seed+1);
    }
    
    private void log(String s){
	output.add(s);
	System.out.println(s);
    }
    
    // -------------------------------
    //		AUTO TEST
    // -------------------------------
    
    private static final int MEASUREMENT_POINTS = 10;
    
    private ArrayList<Long> delayVAC;
    private ArrayList<Long> delayCOST;
    private ArrayList<Long> delayPedVAC;
    private ArrayList<Long> delayPedCOST;
    private ArrayList<Float> accelVAC;
    private ArrayList<Float> accelCOST;
    private ArrayList<Float> posaccelVAC;
    private ArrayList<Float> posaccelCOST;
    
    private boolean autotest_algo = false;
    
    private void autoTestStart(){
	reset(); main.reset();
	useAutoTester = true; autoTest = true;
	log("");
	log("AUTO-TESTER AUTOMATIC TESTER");
	autotest_algo = false;
	renderer.renderWorld.sliderSimSpeed.state = renderer.renderWorld.sliderSimSpeed.max;
	// Random seed selector
	Random rand = new Random();
	seed = rand.nextLong();
	main.rand_uniform.setSeed(seed);
	main.rand_packetloss.setSeed(seed+1);
	// No algo set, lets test the first one
	String str = "Cars: "+((int)renderer.renderWorld.sliderCars.state)+"%";
        str += " | Autos: "+((int)renderer.renderWorld.sliderAutos.state)+"%";
        str += " | Loss: "+((int)renderer.renderWorld.sliderPacketLoss.state)+"%";
        str += " | Peds: "+((int)renderer.renderWorld.sliderPeds.state)+"%";
        log(str);
        str = "Bias";
        str += " | U: "+((int)renderer.renderWorld.sliderBiasU.state)+"%";
        str += " | L: "+((int)renderer.renderWorld.sliderBiasL.state)+"%";
        str += " | D: "+((int)renderer.renderWorld.sliderBiasD.state)+"%";
        str += " | R: "+((int)renderer.renderWorld.sliderBiasR.state)+"%";
        log(str);
        str = "Seed: " + seed;
        str += " | Platoon: "+Car.usePlatooning;
	log(str);
    }
    
    private void autoTestReset(){
	delayVAC = new ArrayList<>();
	delayCOST = new ArrayList<>();
	delayPedVAC = new ArrayList<>();
	delayPedCOST = new ArrayList<>();
	accelVAC = new ArrayList<>();
	accelCOST = new ArrayList<>();
	posaccelVAC = new ArrayList<>();
	posaccelCOST = new ArrayList<>();
    }
    
    private void testAutoTest(){
	// Do our testing
	if( autotest_algo )
	    testThis("vac");
	else
	    testThis("cost_auto_pl");
	// Are we still testing?
	if(main.carCount < TEST_CARCOUNT)
	    return;
	// Else restart our testing and save the current results
	useAutoTester = true;
	if(autotest_algo){
	    log("VAC");
	    delayVAC.add(main.timeSpent);
	    delayPedVAC.add(main.timeSpentPed);
	    accelVAC.add(main.accel);
	    posaccelVAC.add(main.posaccel);
	    // Create new seed
	    Random rand = new Random();
	    seed = rand.nextLong();
	    log("Seed: " + seed+ " | Test count: "+delayVAC.size()+"/"+MEASUREMENT_POINTS);
	}
	else{
	    log("COST_AUTO_PL");
	    delayCOST.add(main.timeSpent);
	    delayPedCOST.add(main.timeSpentPed);
	    accelCOST.add(main.accel);
	    posaccelCOST.add(main.posaccel);
	}
	main.rand_uniform.setSeed(seed);
	main.rand_packetloss.setSeed(seed+1);
	autotest_algo = !autotest_algo;
	main.reset();
	// Check whether we are done here
	if(delayVAC.size() >= MEASUREMENT_POINTS){
	    // Display our results
	    log("AUTO-TESTER AUTO TESTING RESULTS");
	    log("Delay COST: "+UnitConv.loopsToDispString(avgLongs(delayCOST)));
	    log("Delay VAC: "+UnitConv.loopsToDispString(avgLongs(delayVAC)));
	    log("Ped delay COST: "+UnitConv.loopsToDispString(avgLongs(delayPedCOST)));
	    log("Ped delay VAC: "+UnitConv.loopsToDispString(avgLongs(delayPedVAC)));
	    log("Accel COST: "+UnitConv.accelToDispString(avgFloats(accelCOST)));
	    log("Accel VAC: "+UnitConv.accelToDispString(avgFloats(accelVAC)));
	    log("Pos accel COST: "+UnitConv.accelToDispString(avgFloats(posaccelCOST)));
	    log("Pos accel VAC: "+UnitConv.accelToDispString(avgFloats(posaccelVAC)));
	    log("");
	    reset();
	    renderer.renderWorld.sliderSimSpeed.state = renderer.renderWorld.sliderSimSpeed.min;
	}
    }
    
    // -------------------------------
    //		HELPER FUNCS
    // -------------------------------
    
    public static long avgLongs(ArrayList<Long> list){
	long sum = 0;
	for( int i = 0; i < list.size(); i++)
	    sum += list.get(i);
	return sum/list.size();
    }
    
    public static float avgFloats(ArrayList<Float> list){
	float sum = 0;
	for( int i = 0; i < list.size(); i++)
	    sum += list.get(i);
	return sum/list.size();
    }
}
