
package use_trafficintersection.algorithm;

import java.awt.Color;
import java.util.ArrayList;

/**
 * @author Sjoerd Westendorp
 */

public class Lights {
    public static Main main;
    
    public static final boolean USE_WEIGHTED_LANES = false;
    
    // CONSTANTS
    public static final int ERROR = 0;
    public static final int GREEN = 1;
    public static final int YELLOW = 2;
    public static final int RED = 3;
    public static final int SPECIAL = 4;
    public static final int DEFAULT = RED;
    
    public static final int DRAWSIZE = 6;
    public static final int DRAWWSIZE = 2;
    
    public static final int LIGHT_NONE = -1;
    public static final int LIGHT_WD = 1;
    public static final int LIGHT_WU = 2;
    public static final int LIGHT_WL = 3;
    public static final int LIGHT_WR = 4;
    public static final int LIGHT_D = 11;
    public static final int LIGHT_U = 12;
    public static final int LIGHT_L = 13;
    public static final int LIGHT_R = 14;
    public static final int LIGHT_RD = 21;
    public static final int LIGHT_RU = 22;
    public static final int LIGHT_RL = 23;
    public static final int LIGHT_RR = 24;
    
    // The Great System State list
    public ArrayList<int[]> systemStateList = new ArrayList<>();
    public ArrayList<Float> systemStateCostList = new ArrayList<>();
    public ArrayList<Float> systemStateExpectedCostList = new ArrayList<>();
    public int currentSystemState = -1;
    public int nextSystemState = -1;
    public int expectedSystemState = -1;
    
    // Costs
    public static final float CAR_C = 1.0f/150.0f;
    public static final float PED_C = 1.0f/185.0f;
    public static final float CAR_P = 10000.0f;
    public static final float PED_P = 10000.0f;
    public static final float CAR_T = 60.0f*60.0f;
    public static final float PED_T = 60.0f*30.0f;
    
    public static final int TIME_SKIPABLE = 60;
    public static final int TIME_EMPTY = 180;
    
    // renderWolrd: access directly
    // Else: access useing getState() and setState()
    // Lights
    public int LightU,LightD,LightL,LightR;
    // Right lights
    public int LightRU,LightRD,LightRL,LightRR;
    // Walking Lights
    public int LightWU,LightWD,LightWL,LightWR;
    
    // Lane costs
    public float CostU,CostD,CostL,CostR;
    public float CostRU,CostRD,CostRL,CostRR;
    public float CostWU,CostWD,CostWL,CostWR;
    
    // Lane time booleans
    public boolean BoolU,BoolD,BoolL,BoolR;
    public boolean BoolRU,BoolRD,BoolRL,BoolRR;
    public boolean BoolWU,BoolWD,BoolWL,BoolWR;
    
    // Lane time empty booleans
    public boolean EBoolU,EBoolD,EBoolL,EBoolR;
    public boolean EBoolRU,EBoolRD,EBoolRL,EBoolRR;
    public boolean EBoolWU,EBoolWD,EBoolWL,EBoolWR;
    
    // Lane time
    public int TimeU,TimeD,TimeL,TimeR;
    public int TimeRU,TimeRD,TimeRL,TimeRR;
    public int TimeWU,TimeWD,TimeWL,TimeWR;
    
    // Lane empty time
    public int ETimeU,ETimeD,ETimeL,ETimeR;
    public int ETimeRU,ETimeRD,ETimeRL,ETimeRR;
    public int ETimeWU,ETimeWD,ETimeWL,ETimeWR;
    
    public Lights(){
        LightU = DEFAULT;
        LightD = DEFAULT;
        LightL = DEFAULT;
        LightR = DEFAULT;
        
        LightWU = DEFAULT;
        LightWD = DEFAULT;
        LightWL = DEFAULT;
        LightWR = DEFAULT;
	
	LightRU = DEFAULT;
        LightRD = DEFAULT;
        LightRL = DEFAULT;
        LightRR = DEFAULT;
	
	// All pedestrians green, all car lights red.
	systemStateList.add(new int[]{LIGHT_WD,LIGHT_WU,LIGHT_WL,LIGHT_WR});
	
	// 4 rotations of one right turn car lane green, 2 pedestrian routes green.
	systemStateList.add(new int[]{LIGHT_RU,LIGHT_WR,LIGHT_WD});
	systemStateList.add(new int[]{LIGHT_RR,LIGHT_WD,LIGHT_WL});
	systemStateList.add(new int[]{LIGHT_RD,LIGHT_WL,LIGHT_WU});
	systemStateList.add(new int[]{LIGHT_RL,LIGHT_WU,LIGHT_WR});
	
	// 4 rotations of two right turn car lane green, 1 pedestrian routes green.
	systemStateList.add(new int[]{LIGHT_RU,LIGHT_RR,LIGHT_WD});
	systemStateList.add(new int[]{LIGHT_RR,LIGHT_RD,LIGHT_WL});
	systemStateList.add(new int[]{LIGHT_RD,LIGHT_RL,LIGHT_WU});
	systemStateList.add(new int[]{LIGHT_RL,LIGHT_RU,LIGHT_WR});
	
	// All right turn car lanes green, all pedestrians red.
	systemStateList.add(new int[]{LIGHT_RL,LIGHT_RU,LIGHT_RR,LIGHT_RD});
	
	// 4 rotations of one left and straight car lane green, 
	// one right turn lane green and one pedestrian route green.
	algoFixed_systemStateLoc = systemStateList.size();
	algoVac_systemStateLoc = systemStateList.size();
	systemStateList.add(new int[]{LIGHT_U,LIGHT_WL,LIGHT_RR});
	systemStateList.add(new int[]{LIGHT_R,LIGHT_WU,LIGHT_RD});
	systemStateList.add(new int[]{LIGHT_D,LIGHT_WR,LIGHT_RL});
	systemStateList.add(new int[]{LIGHT_L,LIGHT_WD,LIGHT_RU});
	
	// 4 rotations of one left and straight car lane green, 
	// two right turn lane green and all pedestrian routes red.
	systemStateList.add(new int[]{LIGHT_U,LIGHT_RU,LIGHT_RR});
	systemStateList.add(new int[]{LIGHT_R,LIGHT_RR,LIGHT_RD});
	systemStateList.add(new int[]{LIGHT_D,LIGHT_RD,LIGHT_RL});
	systemStateList.add(new int[]{LIGHT_L,LIGHT_RL,LIGHT_RU});
	
	for(int i = 0; i < systemStateList.size(); i++) {
	    systemStateCostList.add((float) i);
	    systemStateExpectedCostList.add((float) i);
	}
    }
    
    public void reset(){
	// Reset lights
	setAllRed();
	// basic algo
	nextGreenState = 0;
	nextGreenStateWiseState = 0;
	// algo
	currentSystemState = -1;
	nextSystemState = -1;
	expectedSystemState = -1;
	// algo fixed
	algoFixed_timer = 0;
	algoFixed_cycleState = 0;
	algoFixed_cycleSection = 0;
	// algo cost
	algoCost_timer = 0;
	algoCost_stage = 0;
	resetLaneCosts();
	for(int i = 0; i < systemStateList.size(); i++){
	    systemStateCostList.set(i, 0.0f);
	    systemStateExpectedCostList.set(i, 0.0f);
	}
	resetLaneTimes();
	resetLaneBools();
	// algo vac
	algoVac_timer = 0;
	algoVac_stage = 0;
	algoVac_cycleState = 0;
    }
    
    // -----------------------------------------
    //	    BASIC ALGORITHMS
    // -----------------------------------------
    
    public void disco(){
        LightU = nextState(LightU);
        LightD = nextState(LightD);
        LightL = nextState(LightL);
        LightR = nextState(LightR);
        
        LightWU = nextState(LightWU);
        LightWD = nextState(LightWD);
        LightWL = nextState(LightWL);
        LightWR = nextState(LightWR);
	
	LightRU = nextState(LightRU);
        LightRD = nextState(LightRD);
        LightRL = nextState(LightRL);
        LightRR = nextState(LightRR);
    }
    
    private int nextGreenState = 0;
    public void nextGreen(){
	nextGreenState++;
	if(nextGreenState == 0){
	    setState(LIGHT_L,RED);
	    setState(LIGHT_RL,RED);
	    
	    setState(LIGHT_WU,GREEN);
	    setState(LIGHT_WD,GREEN);
	    setState(LIGHT_WL,GREEN);
	    setState(LIGHT_WR,GREEN);
	}
	else if(nextGreenState == 1){
	    setState(LIGHT_WU,RED);
	    setState(LIGHT_WD,RED);
	    setState(LIGHT_WL,RED);
	    setState(LIGHT_WR,RED);
	    
	    setState(LIGHT_U,GREEN);
	    setState(LIGHT_RU,GREEN);
	}
	else if(nextGreenState == 2){
	    setState(LIGHT_U,RED);
	    setState(LIGHT_RU,RED);
	    
	    setState(LIGHT_R,GREEN);
	    setState(LIGHT_RR,GREEN);
	}
	else if(nextGreenState == 3){
	    setState(LIGHT_R,RED);
	    setState(LIGHT_RR,RED);
	    
	    setState(LIGHT_D,GREEN);
	    setState(LIGHT_RD,GREEN);
	}
	else if(nextGreenState == 4){
	    setState(LIGHT_D,RED);
	    setState(LIGHT_RD,RED);
	    
	    setState(LIGHT_L,GREEN);
	    setState(LIGHT_RL,GREEN);
	    
	    nextGreenState = -1;
	}
    }
    
    private int nextGreenStateWiseState = 0;
    public void nextGreenStateWise(){
	setSystemState(nextGreenStateWiseState);
	
	nextGreenStateWiseState++;
	if(nextGreenStateWiseState >= systemStateList.size())
	    nextGreenStateWiseState = 0;
    }
    
    private static int nextState(int state){
        switch(state){
            case GREEN:
                return YELLOW;
            case YELLOW:
                return RED;
            case RED:
                return GREEN;
            default:
                return ERROR;
        }
    }
    
    // -----------------------------------------
    //	    SYSTEM STATE
    // -----------------------------------------
    
    public void setSystemState(int systemstate){
	currentSystemState = systemstate;
	nextSystemState = -1;
	int[] arr = systemStateList.get(currentSystemState);
	
	setAllRed();
	for(int i = 0; i < arr.length; i++)
	    setState(arr[i],GREEN);
    }
    
    public void setAllRed(){
	setState(LIGHT_WU,RED);
	setState(LIGHT_WD,RED);
	setState(LIGHT_WL,RED);
	setState(LIGHT_WR,RED);
	
	setState(LIGHT_D,RED);
	setState(LIGHT_RD,RED);
	setState(LIGHT_U,RED);
	setState(LIGHT_RU,RED);
	setState(LIGHT_L,RED);
	setState(LIGHT_RL,RED);
	setState(LIGHT_R,RED);
	setState(LIGHT_RR,RED);
	
    }
    
    public void setTransitionState0(int transstate){
	nextSystemState = transstate;
	int[] next = systemStateList.get(nextSystemState);
	int[] cur;
	if(currentSystemState >= 0)
	    cur = systemStateList.get(currentSystemState);
	else
	    cur = new int[]{};
	
	int light;
	for(int i = 0; i < cur.length; i++){
	    light = cur[i];
	    if( !isIntInIntArray(next,light) && !isPedLight(light)){
		setState(light,YELLOW);
	    }
	}
    }
    
    public void setTransitionState1(int transstate){
	nextSystemState = transstate;
	int[] next = systemStateList.get(nextSystemState);
	int[] cur;
	if(currentSystemState >= 0)
	    cur = systemStateList.get(currentSystemState);
	else
	    cur = new int[]{};
	
	int light;
	for(int i = 0; i < cur.length; i++){
	    light = cur[i];
	    if( !isIntInIntArray(next,light) ){
		setState(light,RED);
	    }
	}
    }
    
    // -----------------------------------------
    //	    ACTUAL ALGORITHMS
    // -----------------------------------------
    
    private static final int ALGO_FIXED_TIMER = 120;
    private int algoFixed_timer = 0;
    private int algoFixed_cycleState = 0;
    private int algoFixed_cycleSection = 0;
    private final int algoFixed_systemStateLoc;
    public void algoFixed(){
	// Fixed time, fixed order of system states etc
	
	// Timer
	if(algoFixed_timer > 0){
	    algoFixed_timer--; 
	    return;
	}
	algoFixed_timer = ALGO_FIXED_TIMER;
	
	// What part of the cycle are we in?
	if( algoFixed_cycleSection == 0){
	    algoFixed_cycleSection++;
	    // Get our next state
	    if(algoFixed_cycleState > 3)
		algoFixed_cycleState = 0;
	    
	    nextSystemState = algoFixed_systemStateLoc+algoFixed_cycleState;
	    
	    algoFixed_cycleState++;
	    // Transition part one
	    setTransitionState0(nextSystemState);
	}
	else if( algoFixed_cycleSection == 1){
	    algoFixed_cycleSection++;
	    // Transition part two
	    
	    setTransitionState1(nextSystemState);
	}
	else if( algoFixed_cycleSection == 2){
	    algoFixed_cycleSection++;
	    // Set state
	    
	    setSystemState(nextSystemState);
	}
	else{
	    algoFixed_cycleSection = 0;
	    // Wait for an extra cycle
	}
    }
    
    private static final int ALGO_COST_TIMER_TRANS0 = 70;
    private static final int ALGO_COST_TIMER_TRANS1 = 90;
    private static final int ALGO_COST_TIMER_STATE = 500;
    private static final int ALGO_COST_TIMER_STATE_MIN = 100;
    private static final int ALGO_COST_ANNOUCETIME = ALGO_COST_TIMER_TRANS0+ALGO_COST_TIMER_TRANS1;
    private static final int ALGO_COST_ANNOUCETIME_EXPECTED = ALGO_COST_TIMER_STATE_MIN+2*ALGO_COST_TIMER_TRANS0+2*ALGO_COST_TIMER_TRANS1;
    private int algoCost_timer = 0;
    private int algoCost_stage = 0;
    public void algoCost(boolean useAutos, boolean usePacketLoss){
	updateLaneTimes();
	
	// Check packetloss setting (purely for peformance)
	if(usePacketLoss && Main.renderer.getPacketLoss() < 1.0f)
	    usePacketLoss = false;
	
	// Timer
	if(algoCost_timer > 0){
	    algoCost_timer--; 
	    if(algoCost_stage != 0)
		return;
	    // Early red system
	    if(!isCurrentStateEmpty())
		return;
	    if(algoCost_timer > ALGO_COST_TIMER_STATE - ALGO_COST_TIMER_STATE_MIN)
		return;
	}
	
	// What part of the cycle are we in?
	if( algoCost_stage == 0){
	    algoCost_stage++;
	    algoCost_timer = ALGO_COST_TIMER_TRANS0;
	    // Get our next state
	    nextSystemState = algoCostGetNextSystemState(useAutos,usePacketLoss);
	    // Transition part one
	    setTransitionState0(nextSystemState);
	    // Annouce this to the autonomous cars
	    if(useAutos){
		announceState(nextSystemState,ALGO_COST_ANNOUCETIME,usePacketLoss);
		// Get expected next system state
		expectedSystemState = algoCostGetExpectedSystemState(useAutos,usePacketLoss);
		// Then announce this to the autonomous cars
		announceExpectedState(expectedSystemState,ALGO_COST_ANNOUCETIME_EXPECTED,usePacketLoss);
	    }
	}
	else if( algoCost_stage == 1){
	    algoCost_stage++;
	    algoCost_timer = ALGO_COST_TIMER_TRANS1;
	    // Transition part two
	    setTransitionState1(nextSystemState);
	}
	else{
	    algoCost_stage = 0;
	    algoCost_timer = ALGO_COST_TIMER_STATE;
	    // Set state
	    setSystemState(nextSystemState);
	}
    }
    
    private int algoCostGetNextSystemState(boolean useAutos, boolean usePacketLoss){
	// For all lanes, calculate cost
	calcLaneCosts(usePacketLoss);
	// For all system states, calculate cost reduction
	int arr[]; float cost;
	for(int i = 0; i < systemStateList.size(); i++){
	    cost = 0;
	    arr = systemStateList.get(i);
	    // Get cost
	    if( useAutos )
		cost += getSystemStateLanesCost(arr);
	    cost += getSystemStateTimeCost(arr);
	    // Place cost in list
	    systemStateCostList.set(i, cost);
	}
	// Get highest cost reduction state
	int highestIndex = 0;
	float highestCost = systemStateCostList.get(0), compcost;
	for(int i = 1; i < systemStateCostList.size(); i++){
	    compcost = systemStateCostList.get(i);
	    if( compcost > highestCost ){
		highestCost = compcost;
		highestIndex = i;
	    }
	}
	// Return this state
	return highestIndex;
    }
    
    private int algoCostGetExpectedSystemState(boolean useAutos, boolean usePacketLoss){
	// We assume that the next system state will clear its lanes
	// Now we just need to recalculate the costs
	recalcLaneCosts(nextSystemState);
	// For all system states, calculate cost reduction
	int arr[]; float cost;
	for(int i = 0; i < systemStateList.size(); i++){
	    cost = 0;
	    arr = systemStateList.get(i);
	    // Get cost
	    if( useAutos )
		cost += getSystemStateLanesCost(arr);
	    cost += getSystemStateTimeExpectedCost(arr,nextSystemState);
	    // Place cost in list
	    systemStateExpectedCostList.set(i, cost);
	}
	// Get highest cost reduction state
	int highestIndex = 0;
	float highestCost = systemStateExpectedCostList.get(0), compcost;
	for(int i = 1; i < systemStateExpectedCostList.size(); i++){
	    compcost = systemStateExpectedCostList.get(i);
	    if( compcost > highestCost ){
		highestCost = compcost;
		highestIndex = i;
	    }
	}
	// Return this state
	return highestIndex;
    }
    
    private void announceState(int state, int time, boolean usePacketLoss){
	int lights[] = systemStateList.get(state);
	// Tell each auto this
	Car car;
	boolean correctLight;
	for(int i = 0; i < main.cars.size();i++){
	    // Vars
	    car = main.cars.get(i);
	    // is it a auto?
	    if(!car.isAuto)
		continue;
	    // Is it before the intersection
	    if(car.state != Car.STATE_PRE_JUNCTION)
		continue;
	    // Check if it is indeed nearing a relevant light
	    correctLight = false;
	    for(int j = 0; j < lights.length; j++){
		if( car.light == lights[j] ){
		    correctLight = true;
		    break;
		}
	    }
	    if(!correctLight)
		continue;
	    // Did we lose the packet?
	    if(usePacketLoss && isPacketLost())
		continue;
	    // Lets tell it what to do
	    car.tellNextGreenWave(time);
	}
    }
    
    private void announceExpectedState(int state, int time, boolean usePacketLoss){
	int lights[] = systemStateList.get(state);
	// Tell each auto this
	Car car;
	boolean correctLight;
	for(int i = 0; i < main.cars.size();i++){
	    // Vars
	    car = main.cars.get(i);
	    // is it a auto?
	    if(!car.isAuto)
		continue;
	    if(car.state != Car.STATE_PRE_JUNCTION)
		continue;
	    // Check if it is indeed nearing a relevant light
	    correctLight = false;
	    for(int j = 0; j < lights.length; j++){
		if( car.light == lights[j] ){
		    correctLight = true;
		    break;
		}
	    }
	    if(!correctLight)
		continue;
	    // Did we lose the packet?
	    if(usePacketLoss && isPacketLost())
		continue;
	    // Lets tell it what to do
	    car.tellExpectedGreenWave(time);
	}
    }
    
    private static final int ALGO_VAC_TIMER_TRANS0 = 70;
    private static final int ALGO_VAC_TIMER_TRANS1 = 90;
    private static final int ALGO_VAC_TIMER_STATE = 600;
    private static final int ALGO_VAC_TIMER_STATE_MIN = 100;
    private int algoVac_timer = 0;
    private int algoVac_cycleState = 0;
    private int algoVac_stage = 0;
    private final int algoVac_systemStateLoc;
    public void algoVac(){
	updateLaneTimes();
	
	// Timer
	if(algoVac_timer > 0){
	    algoVac_timer--; 
	    if(algoVac_stage != 0)
		return;
	    // Early red system
	    if(!isCurrentStateEmpty())
		return;
	    if(algoVac_timer > ALGO_VAC_TIMER_STATE - ALGO_VAC_TIMER_STATE_MIN)
		return;
	}
	
	// What part of the cycle are we in?
	if( algoVac_stage == 0){
	    algoVac_stage++;
	    algoVac_timer = ALGO_VAC_TIMER_TRANS0;
	    // Get our next state
	    nextSystemState = algoVacGetNextSystemState();
	    // Transition part one
	    setTransitionState0(nextSystemState);
	}
	else if( algoVac_stage == 1){
	    algoVac_stage++;
	    algoVac_timer = ALGO_VAC_TIMER_TRANS1;
	    // Transition part two
	    setTransitionState1(nextSystemState);
	}
	else{
	    algoVac_stage = 0;
	    algoVac_timer = ALGO_VAC_TIMER_STATE;
	    // Set state
	    setSystemState(nextSystemState);
	}
    }
  
    private int algoVacGetNextSystemState(){
	boolean skip;
	int loopCount = 0;
	do{
	    loopCount++;
	    // lets first go to the next state
	    algoVac_cycleState++;
	    if(algoVac_cycleState >= 4)
		algoVac_cycleState = 0;
	    // Check if we need to skip it
	    skip = isStateSkipable(algoVac_systemStateLoc+algoVac_cycleState);
	} while(skip && loopCount < 4);
	
	// If we went full circle, just use the next one
	if(loopCount >= 4){
	    algoVac_cycleState++;
	    if(algoVac_cycleState >= 4)
		algoVac_cycleState = 0;
	}
	
	// return our new state
	return algoVac_systemStateLoc+algoVac_cycleState;
    }
    
    private boolean isStateSkipable(int state){
	int[] lights = systemStateList.get(state);
	int light, time;
	
	// Try to find a non-skipable lane
	for( int i = 0; i < lights.length; i++){
	    light = lights[i];
	    time = getLaneTime(light);
	    
	    if(time >= TIME_SKIPABLE)
		return false;
	}
	
	// No non-skipable lanes found
	return true;
    }
    
    // -----------------------------------------
    //	    EMPTY LANE CALCULATIONS
    // -----------------------------------------
    
    private boolean isCurrentStateEmpty(){
	// Catch special case
	if(currentSystemState == -1)
	    return false; // has to be false, true would cause madness
	
	int[] lights = systemStateList.get(currentSystemState);
	int light, time;
	
	// Try to find a non-empty lane
	for( int i = 0; i < lights.length; i++){
	    light = lights[i];
	    time = getLaneETime(light);
	    
	    if(time < TIME_EMPTY)
		return false;
	}
	
	// No non-empty lanes found
	return true;
    }
    
    private int getLaneETime(int light){
	switch(light){
	    case LIGHT_WD:
		return ETimeWD;
	    case LIGHT_WU:
		return ETimeWU;
	    case LIGHT_WR:
		return ETimeWR;
	    case LIGHT_WL:
		return ETimeWL;
	    case LIGHT_D:
		return ETimeD;
	    case LIGHT_U:
		return ETimeU;
	    case LIGHT_R:
		return ETimeR;
	    case LIGHT_L:
		return ETimeL;
	    case LIGHT_RD:
		return ETimeRD;
	    case LIGHT_RU:
		return ETimeRU;
	    case LIGHT_RR:
		return ETimeRR;
	    case LIGHT_RL:
		return ETimeRL;
	    default:
		return -1;
	}
    }
    
    public void setLaneEBool(int light){
	boolean state = true;
	switch(light){
	    case LIGHT_WD:
		EBoolWD = state; break;
	    case LIGHT_WU:
		EBoolWU = state; break;
	    case LIGHT_WR:
		EBoolWR = state; break;
	    case LIGHT_WL:
		EBoolWL = state; break;
	    case LIGHT_D:
		EBoolD = state; break;
	    case LIGHT_U:
		EBoolU = state; break;
	    case LIGHT_R:
		EBoolR = state; break;
	    case LIGHT_L:
		EBoolL = state; break;
	    case LIGHT_RD:
		EBoolRD = state; break;
	    case LIGHT_RU:
		EBoolRU = state; break;
	    case LIGHT_RR:
		EBoolRR = state; break;
	    case LIGHT_RL:
		EBoolRL = state; break;
	    default:
	}
    }
    
    // -----------------------------------------
    //	    TIME CALCULATIONS
    // -----------------------------------------
    
    private void resetLaneTimes(){
	TimeU = 0; 
	TimeD = 0;
	TimeL = 0;
	TimeR = 0;
	TimeRU = 0;
	TimeRD = 0;
	TimeRL = 0;
	TimeRR = 0;
	TimeWU = 0;
	TimeWD = 0;
	TimeWL = 0;
	TimeWR = 0;
	ETimeU = 0; 
	ETimeD = 0;
	ETimeL = 0;
	ETimeR = 0;
	ETimeRU = 0;
	ETimeRD = 0;
	ETimeRL = 0;
	ETimeRR = 0;
	ETimeWU = 0;
	ETimeWD = 0;
	ETimeWL = 0;
	ETimeWR = 0;
    }
    
    private void resetLaneBools(){
	BoolU = false; 
	BoolD = false;
	BoolL = false;
	BoolR = false;
	BoolRU = false;
	BoolRD = false;
	BoolRL = false;
	BoolRR = false;
	BoolWU = false;
	BoolWD = false;
	BoolWL = false;
	BoolWR = false;
	
	EBoolU = false; 
	EBoolD = false;
	EBoolL = false;
	EBoolR = false;
	EBoolRU = false;
	EBoolRD = false;
	EBoolRL = false;
	EBoolRR = false;
	EBoolWU = false;
	EBoolWD = false;
	EBoolWL = false;
	EBoolWR = false;
    }
    
    private float getSystemStateTimeCost(int[] lights){
	int light, time;
	float cost = 0.0f;
	
	for( int i = 0; i < lights.length; i++){
	    light = lights[i];
	    time = getLaneTime(light);
	    
	    if(isPedLight(light)){
		cost += PED_C*(float)time;
		if(time > PED_T)
		    cost += PED_P;
	    }
	    else if(USE_WEIGHTED_LANES && isLeftAndForwardLight(light)){
		cost += 2*CAR_C*(float)time;
		if(time > CAR_T)
		    cost += CAR_P;
	    }
	    else{
		cost += CAR_C*(float)time;
		if(time > CAR_T)
		    cost += CAR_P;
	    }
	}
	return cost;
    }
    
    private float getSystemStateTimeExpectedCost(int[] lights, int subtract_systemstate){
	int[] subtract_lights = systemStateList.get(subtract_systemstate);
	
	int light, time;
	float cost = 0.0f;
	for( int i = 0; i < lights.length; i++){
	    light = lights[i];
	    time = getLaneTime(light);
	    
	    // if light is in the next system state, then skip it
	    boolean skip = false;
	    for(int k = 0; k < subtract_lights.length; k++){
		if(subtract_lights[k] == light){
		    skip = true;
		    break;
		}
	    }
	    if(skip)
		continue;
	    
	    // Else do the usual thing
	    if(isPedLight(light)){
		cost += PED_C*(float)time;
		if(time > PED_T)
		    cost += PED_P;
	    }
	    else if(USE_WEIGHTED_LANES && isLeftAndForwardLight(light)){
		cost += 2*CAR_C*(float)time;
		if(time > CAR_T)
		    cost += CAR_P;
	    }
	    else{
		cost += CAR_C*(float)time;
		if(time > CAR_T)
		    cost += CAR_P;
	    }
	}
	return cost;
    }
    
    private int getLaneTime(int light){
	switch(light){
	    case LIGHT_WD:
		return TimeWD;
	    case LIGHT_WU:
		return TimeWU;
	    case LIGHT_WR:
		return TimeWR;
	    case LIGHT_WL:
		return TimeWL;
	    case LIGHT_D:
		return TimeD;
	    case LIGHT_U:
		return TimeU;
	    case LIGHT_R:
		return TimeR;
	    case LIGHT_L:
		return TimeL;
	    case LIGHT_RD:
		return TimeRD;
	    case LIGHT_RU:
		return TimeRU;
	    case LIGHT_RR:
		return TimeRR;
	    case LIGHT_RL:
		return TimeRL;
	    default:
		return -1;
	}
    }
    
    private void updateLaneTimes(){
	/*
	* It was already way to late to change it, 
	* when I realised this function had to be created.
	* Never create a set of objects,
	* without having an array for them to inject.
	*/
	if(BoolWD){ TimeWD++;  }
	else{ TimeWD = 0;}
	if(BoolWU){ TimeWU++; }
	else{ TimeWU = 0; }
	if(BoolWR){ TimeWR++; }
	else{ TimeWR = 0; }
	if(BoolWL){ TimeWL++; }
	else{ TimeWL = 0;  }
	if(BoolRD){ TimeRD++; }
	else{ TimeRD = 0; }
	if(BoolRU){ TimeRU++;  }
	else{ TimeRU = 0;  }
	if(BoolRR){ TimeRR++; }
	else{ TimeRR = 0; }
	if(BoolRL){ TimeRL++; }
	else{ TimeRL = 0; }
	if(BoolD){ TimeD++;}
	else{ TimeD = 0;  }
	if(BoolU){ TimeU++;  }
	else{ TimeU = 0; }
	if(BoolR){ TimeR++; }
	else{ TimeR = 0; }
	if(BoolL){ TimeL++;  }
	else{ TimeL = 0; }
	
	if(EBoolWD){  ETimeWD = 0; }
	else{  ETimeWD++; }
	if(EBoolWU){ETimeWU = 0; }
	else{  ETimeWU++;}
	if(EBoolWR){ ETimeWR = 0; }
	else{  ETimeWR++; }
	if(EBoolWL){  ETimeWL = 0; }
	else{ ETimeWL++; }
	if(EBoolRD){  ETimeRD = 0; }
	else{  ETimeRD++; }
	if(EBoolRU){  ETimeRU = 0; }
	else{  ETimeRU++; }
	if(EBoolRR){ ETimeRR = 0; }
	else{ ETimeRR++; }
	if(EBoolRL){ ETimeRL = 0; }
	else{  ETimeRL++; }
	if(EBoolD){ ETimeD = 0; }
	else{  ETimeD++; }
	if(EBoolU){  ETimeU = 0; }
	else{  ETimeU++; }
	if(EBoolR){  ETimeR = 0; }
	else{  ETimeR++; }
	if(EBoolL){  ETimeL = 0; }
	else{  ETimeL++; }
	resetLaneBools();
    }
    
    public void setLaneBool(int light){
	boolean state = true;
	switch(light){
	    case LIGHT_WD:
		BoolWD = state; break;
	    case LIGHT_WU:
		BoolWU = state; break;
	    case LIGHT_WR:
		BoolWR = state; break;
	    case LIGHT_WL:
		BoolWL = state; break;
	    case LIGHT_D:
		BoolD = state; break;
	    case LIGHT_U:
		BoolU = state; break;
	    case LIGHT_R:
		BoolR = state; break;
	    case LIGHT_L:
		BoolL = state; break;
	    case LIGHT_RD:
		BoolRD = state; break;
	    case LIGHT_RU:
		BoolRU = state; break;
	    case LIGHT_RR:
		BoolRR = state; break;
	    case LIGHT_RL:
		BoolRL = state; break;
	    default:
	}
    }
    
    // -----------------------------------------
    //	    COST CALCULATIONS
    // -----------------------------------------
    
    private float getSystemStateLanesCost(int[] lights){
	int light;
	float cost = 0.0f;
	for( int i = 0; i < lights.length; i++){
	    light = lights[i];
	    cost += getLaneCost(light);
	}
	return cost;
    }
    
    private void calcLaneCosts(boolean usePacketLoss){
	// Reset lane costs
	resetLaneCosts();
	// Go over all the autos and add them to the lane costs
	//float automultiplier = 1/(Main.renderer.renderWorld.sliderAutos.state/100.0f)+0.01f;
	float automultiplier = 1.0f;
	float carmultiplier = 1.0f;
	Car car;
	int light;
	for(int i = 0; i < main.cars.size(); i++) {
	    car = main.cars.get(i);
	    if(!car.isAuto)
		continue;
	    if(car.state != Car.STATE_PRE_JUNCTION)
		continue;
	    if(usePacketLoss && isPacketLost())
		continue;
	    light = dirToLight(car.dirStart,car.dirRel);
	    addLaneCost(light,automultiplier);
	    // Check car in front
	    if(car.carInFront != null
		&& !car.carInFront.isAuto // All autos report their existance themselves
		&& car.carInFrontDis < Car.CLOSE_TO_CAR
		    )
		addLaneCost(light,carmultiplier);
	    // Check car behind
	    if(car.carBehind != null && !car.carBehind.isAuto)
		addLaneCost(light,carmultiplier);
	}
    }
    
    private void recalcLaneCosts(int subtract_systemstate){
	int[] lights = systemStateList.get(subtract_systemstate);
	
	int light;
	for( int i = 0; i < lights.length; i++){
	    light = lights[i];
	    // Set cost of lane to zero
	    setLaneCost(light,0.0f);
	}
    }
    
    private float getLaneCost(int light){
	switch(light){
	    case LIGHT_WD:
		return CostWD;
	    case LIGHT_WU:
		return CostWU;
	    case LIGHT_WR:
		return CostWR;
	    case LIGHT_WL:
		return CostWL;
	    case LIGHT_D:
		return CostD;
	    case LIGHT_U:
		return CostU;
	    case LIGHT_R:
		return CostR;
	    case LIGHT_L:
		return CostL;
	    case LIGHT_RD:
		return CostRD;
	    case LIGHT_RU:
		return CostRU;
	    case LIGHT_RR:
		return CostRR;
	    case LIGHT_RL:
		return CostRL;
	    default:
		return -1;
	}
    }
    
    private void addLaneCost(int light, float state){
	switch(light){
	    case LIGHT_WD:
		CostWD += state; break;
	    case LIGHT_WU:
		CostWU += state; break;
	    case LIGHT_WR:
		CostWR += state; break;
	    case LIGHT_WL:
		CostWL += state; break;
	    case LIGHT_D:
		CostD += state; break;
	    case LIGHT_U:
		CostU += state; break;
	    case LIGHT_R:
		CostR += state; break;
	    case LIGHT_L:
		CostL += state; break;
	    case LIGHT_RD:
		CostRD += state; break;
	    case LIGHT_RU:
		CostRU += state; break;
	    case LIGHT_RR:
		CostRR += state; break;
	    case LIGHT_RL:
		CostRL += state; break;
	    default:
	}
    }
    
    private void setLaneCost(int light, float state){
	switch(light){
	    case LIGHT_WD:
		CostWD = state; break;
	    case LIGHT_WU:
		CostWU = state; break;
	    case LIGHT_WR:
		CostWR = state; break;
	    case LIGHT_WL:
		CostWL = state; break;
	    case LIGHT_D:
		CostD = state; break;
	    case LIGHT_U:
		CostU = state; break;
	    case LIGHT_R:
		CostR = state; break;
	    case LIGHT_L:
		CostL = state; break;
	    case LIGHT_RD:
		CostRD = state; break;
	    case LIGHT_RU:
		CostRU = state; break;
	    case LIGHT_RR:
		CostRR = state; break;
	    case LIGHT_RL:
		CostRL = state; break;
	    default:
	}
    }
    
    private void resetLaneCosts(){
	CostU = 0; 
	CostD = 0;
	CostL = 0;
	CostR = 0;
	CostRU = 0;
	CostRD = 0;
	CostRL = 0;
	CostRR = 0;
	CostWU = 0;
	CostWD = 0;
	CostWL = 0;
	CostWR = 0;
    }
    
    // -----------------------------------------
    //	    LIGHT STATES
    // -----------------------------------------
    
    public static Color getColor(int state){
        switch(state){
            case GREEN:
                return new Color(50,255,50);
            case YELLOW:
                return new Color(255,180,0);
            case RED:
                return Color.RED;
            case SPECIAL:
                return Color.CYAN;
            case ERROR:
                return Color.BLUE;
            default:
                return Color.PINK;
        }
    }
    
    public String getStateString(int light){
	int i = getState(light);
	if( i == RED )
	    return "RED";
	else if( i == GREEN )
	    return "GREEN";
	else if( i == YELLOW )
	    return "YELLOW";
	else if( i == SPECIAL )
	    return "SPECIAL";
	else
	    return "ERROR";
    }
    
    public int getState(int light){
	switch(light){
	    case LIGHT_WD:
		return LightWD;
	    case LIGHT_WU:
		return LightWU;
	    case LIGHT_WR:
		return LightWR;
	    case LIGHT_WL:
		return LightWL;
	    case LIGHT_D:
		return LightD;
	    case LIGHT_U:
		return LightU;
	    case LIGHT_R:
		return LightR;
	    case LIGHT_L:
		return LightL;
	    case LIGHT_RD:
		return LightRD;
	    case LIGHT_RU:
		return LightRU;
	    case LIGHT_RR:
		return LightRR;
	    case LIGHT_RL:
		return LightRL;
	    default:
		return ERROR;
	}
    }
    
    public void setState(int light, int state){
	switch(light){
	    case LIGHT_WD:
		LightWD = state; break;
	    case LIGHT_WU:
		LightWU = state; break;
	    case LIGHT_WR:
		LightWR = state; break;
	    case LIGHT_WL:
		LightWL = state; break;
	    case LIGHT_D:
		LightD = state; break;
	    case LIGHT_U:
		LightU = state; break;
	    case LIGHT_R:
		LightR = state; break;
	    case LIGHT_L:
		LightL = state; break;
	    case LIGHT_RD:
		LightRD = state; break;
	    case LIGHT_RU:
		LightRU = state; break;
	    case LIGHT_RR:
		LightRR = state; break;
	    case LIGHT_RL:
		LightRL = state; break;
	    default:
	}
    }

    public static boolean isPedLight(int light){
	if(light == LIGHT_WU)
	    return true;
	else if(light == LIGHT_WR)
	    return true;
	else if(light == LIGHT_WD)
	    return true;
	else if(light == LIGHT_WL)
	    return true;
	return false;
    }
    
    public static boolean isLeftAndForwardLight(int light){
	if(light == LIGHT_U)
	    return true;
	else if(light == LIGHT_R)
	    return true;
	else if(light == LIGHT_D)
	    return true;
	else if(light == LIGHT_L)
	    return true;
	return false;
    }
    
    // -----------------------------------------
    //	    HELPER FUNCTIONS
    // -----------------------------------------
    
    public boolean isPacketLost(){
	return (main.rand_packetloss.nextFloat() < (Main.renderer.getPacketLoss()/100.0f));
    }
    
    public static boolean isIntInIntArray(int[] systemstate, int light){
	for(int i = 0; i < systemstate.length; i++){
	    if(systemstate[i] == light)
		return true;
	}
	return false;
    }
    
    public int dirToLight(int dirStart,int dirRel){
	if(dirStart == Dir.UP){
	    if(dirRel == Dir.GO_RIGHT)
		return LIGHT_RU;
	    return LIGHT_U;
	}
	else if(dirStart == Dir.LEFT){
	    if(dirRel == Dir.GO_RIGHT)
		return LIGHT_RL;
	    return LIGHT_L;
	}
	else if(dirStart == Dir.DOWN){
	    if(dirRel == Dir.GO_RIGHT)
		return LIGHT_RD;
	    return LIGHT_D;
	}
	else{
	    if(dirRel == Dir.GO_RIGHT)
		return LIGHT_RR;
	    return LIGHT_R;
	}
    }
}