
package use_trafficintersection.algorithm;
import java.util.ArrayList;
import java.util.Random;
import use_trafficintersection.*;

/**
 * @author Sjoerd Westendorp
 */

/*
From here the algorithm and the Renderer interface
The renderer controls the program and calls three functions of Main
- The constructor, allowing the algorithm the setitself up.
- Loop, where the algorithm does its job.
- Draw, where the renderer gets told where the cars are and what else to draw.

BTW: The renderer loves uneven numbers. 
Almost everything gets displayed better if their size is uneven.
*/

public class Main {
    public static final String STANDARD_ALGO = "cost_auto_pl";
    
    // Other objects
    public static Renderer renderer;
    public static Lights lights;
    public static AutoTester autoTester;
    
    // Cars and peds lists
    public ArrayList<Car> cars = new ArrayList<>();
    public ArrayList<Ped> peds = new ArrayList<>();
    
    // Debug value that get displayed by the renderer
    public int debugInt = Renderer.DEBUG_INT_NOSHOW;
    public String debug = "";
    
    // Time spent & accel
    public long carCount = 0;
    public long timeSpent = 0; // Gets displayed by renderer
    public final ArrayList<Long> timeSpentList = new ArrayList<>();
    public float accel = 0; // Gets displayed by renderer
    public final ArrayList<Float> accelList = new ArrayList<>();
    public float posaccel = 0; // Gets displayed by renderer
    public final ArrayList<Float> posaccelList = new ArrayList<>();
    
    public long timeSpentPed = 0; // Gets displayed by renderer
    public final ArrayList<Long> timeSpentListPed = new ArrayList<>();
    public long pedCount = 0;
    
    // Constants
    public static final int PED_SIZE = 3;
    public static final int CAR_WIDTH = Renderer.CAR_MAX_WIDTH;
    public static final int CAR_LENGHT = CAR_WIDTH*3/2;
    
    public static final int PED_TIMER_BASE = 100;
    public static final int PED_TIMER = 100;
    public static final int CAR_TIMER_BASE = 40;
    public static final int CAR_TIMER = 60;
    public static final int CAR_TIMER_RAND = 100;
    
    public static final int TIME_SPENT_LIST_MAX_SIZE = 500;
    
    // Collision exception, throwing it pauses simulation
    public class CollisionException extends Exception{
	String text;
	
	CollisionException(){
	    text = "CollisionException";
	}
	
	CollisionException(String s){
	    text = s;
	}
	
	public String getText(){
	    return text;
	}
    }
    
    // ------------------------------------
    
    public Main(Renderer r,Lights l){
	// Place a reference to renderer and lights into main class.
	// We are going to need this to draw things on screen.
	renderer = r;
	lights = l;
	/*
	    Setup algorithm function
	    Do not use renderer or anything graphical inside this function
	*/
	
	/*
	DEBUG FUNCTIONS
	System.out.println("LANESIZE: "+RenderWorld.LANESIZE);
	System.out.println("SIDESIZE: "+RenderWorld.SIDESIZE);
	System.out.println("JUNCSIZE: "+Renderer.JUNCSIZE);
	System.out.println("LANE_OFFSET: "+Renderer.LANE_OFFSET);
	System.out.println("SIDE_OFFSET: "+Renderer.SIDE_OFFSET);
	*/
    }
    
    public void loop() throws CollisionException{
	/*
	    Have the algorithm calculate and simulate the next step
	    If collision between cars or car and pedestrian occurs, 
	    throw a CollisionException (throw new CollisionException();)
	*/
	if(AutoTester.useAutoTester){
	    autoTester.test();
	}
	else{
	    //lights.algoCost(true,true);
	    autoTester.testThis(STANDARD_ALGO);
	}
	
	carSpawner();
	pedSpawner();
	
        //Car movement
	Car car;
	for(int i = cars.size()-1; i >= 0;  i--){
	    car = cars.get(i);
	    car.step(cars,peds);
	    if(car.isDone){
		addTimeSpent(renderer.loopCount - car.loopExpected,car.accel,car.posaccel);
		cars.remove(i);
	    }
	}
        
        
	Ped ped;
	for(int i = peds.size()- 1; i >= 0; i--){
	    ped = peds.get(i);
	    ped.step();
	    if(ped.isDone){
                addPedTimeSpent(renderer.loopCount - ped.loopExpected);
		peds.remove(i);
	    }
	}
    }
    
    public void draw(){
	/*
	    Draw the vehicles
	    - Drawing the world already has been done
	    - Drawing the top bar will be done after this
	*/
	if(AutoTester.unitTest)
	    debug = "[UnitTest] "+autoTester.unitTestName;
	else if(AutoTester.autoTest)
	    debug = "[AutoTest]";
	else
	    debug = "";
	
	Car car;
	for(int i = 0; i < cars.size(); i++){
	    car = cars.get(i);
	    if(renderer.debugData){
		car.drawForwardCheck(renderer);
	    }
	    renderer.drawCar((int) car.x,(int)  car.y, car.w, car.h, car.c, car.rotation);
	}
	
	Ped ped;
	for(int i = 0; i < peds.size(); i++){
	    ped = peds.get(i);
	    renderer.drawPed(ped.x(), ped.y(), ped.r, ped.c);
	    //renderer.drawCross((int) ped.x,(int)  ped.y);
	}
    }
    
    public void reset(){
	cars.clear();
	peds.clear();
	timeSpentList.clear();
	timeSpent = 0;
	timeSpentListPed.clear();
	timeSpentPed = 0;
	accelList.clear();
	accel = 0;
	posaccelList.clear();
	posaccel = 0;
	lights.reset();
	carCount = 0;
	carSpawner_timer = CAR_TIMER_BASE;
	pedSpawner_timer = PED_TIMER_BASE;
    }
    
    // --------------------------------
    //		    MEASURING
    // --------------------------------
    
    private void addPedTimeSpent(long time){
        timeSpentListPed.add(time);
	pedCount++;
        
        // Keep list at reasonable length
	if(timeSpentListPed.size() > TIME_SPENT_LIST_MAX_SIZE){
	    timeSpentListPed.remove(0);
	}
	
	// Recalc average
	long sum = 0;
	for(int i = 0; i < timeSpentListPed.size(); i ++){
	    sum += timeSpentListPed.get(i);
	}
	timeSpentPed = sum/timeSpentListPed.size();
    }
    
    private void addTimeSpent(long time,float acc,float posacc){
	timeSpentList.add(time);
	accelList.add(acc);
	posaccelList.add(posacc);
	carCount++;
	
	// Keep list at reasonable length
	if(timeSpentList.size() > TIME_SPENT_LIST_MAX_SIZE){
	    timeSpentList.remove(0);
	}
	if(accelList.size() > TIME_SPENT_LIST_MAX_SIZE){
	    accelList.remove(0);
	}
	if(posaccelList.size() > TIME_SPENT_LIST_MAX_SIZE){
	    posaccelList.remove(0);
	}
	
	// Recalc average
	long sum = 0;
	for(int i = 0; i < timeSpentList.size(); i ++){
	    sum += timeSpentList.get(i);
	}
	timeSpent = sum/timeSpentList.size();
	
	float accsum = 0;
	for(int i = 0; i < accelList.size(); i ++){
	    accsum += accelList.get(i);
	}
	accel = accsum/accelList.size();
	
	float posaccsum = 0;
	for(int i = 0; i < posaccelList.size(); i ++){
	    posaccsum += posaccelList.get(i);
	}
	posaccel = posaccsum/posaccelList.size();
    }
    
    // --------------------------------
    //		    SPAWNERS
    // --------------------------------
    
    public Random rand_packetloss = new Random();
    public Random rand_uniform = new Random();
    
    private int carSpawner_timer = CAR_TIMER_BASE;
    private void carSpawner(){
	float spawnrate = renderer.getCars();
	if( spawnrate < 1.0f )
	    return;
	
	carSpawner_timer--;
	if( carSpawner_timer > 0 )
	    return;
	carSpawner_timer = CAR_TIMER_BASE + (CAR_TIMER * (int)(100.0f - spawnrate) ) / 100;
	carSpawner_timer += CAR_TIMER_RAND*nextNonUniformRandom();
	//carSpawner_timer = CAR_TIMER_BASE; 
	//carSpawner_timer += (int) (nextNonUniformRandom()*(CAR_TIMER * (100.0f - spawnrate) ) / 100.0f);
	
	// --- Starting direction ---
	// Get bias
	float f[] = renderer.getBias();
	float biasU = f[0], biasD = f[1], biasL = f[2], biasR = f[3];
	
	// Do the random stuff
	float randf = (float)(rand_uniform.nextFloat() * (biasU+biasD+biasL+biasR));
	int dir_start;
	if( randf < biasU )
	    dir_start = Dir.UP;
	else if( randf < biasU+biasD )
	    dir_start = Dir.DOWN;
	else if( randf < biasU+biasD+biasR )
	    dir_start = Dir.RIGHT;
	else
	    dir_start = Dir.LEFT;
	
	// --- Relative direction --- 
	int rand = (int)(rand_uniform.nextFloat() * 3);
	int dir_rel;
	switch (rand) {
	    case 0:
		dir_rel = Dir.GO_LEFT;
		break;
	    case 1:
		dir_rel = Dir.GO_RIGHT;
		break;
	    default:
		dir_rel = Dir.GO_FORWARD;
		break;
	}
	
	// Is auto
	boolean isAuto = (rand_uniform.nextFloat() < (renderer.getAutos()/100.0f));
	
	// Create car
	Car newcar = new Car(lights,CAR_WIDTH,CAR_LENGHT,dir_start,dir_rel,isAuto,renderer.loopCount);
	
	// Check if empty
	Car compcar;
	for(int i = 0; i < cars.size(); i++){
	    compcar = cars.get(i);
	    if(compcar.dirStart != newcar.dirStart)
		continue; //not starting at the same side
	    if((compcar.dirRel == Dir.GO_RIGHT) != (newcar.dirRel == Dir.GO_RIGHT))
		continue; //not in the same lane
	    if(disSq(compcar.x,compcar.y,newcar.x,newcar.y) > compcar.hitRadius2*compcar.hitRadius2)
		continue; //not close enough
	    return;
	}
	
	// Spawn car
        cars.add(newcar);
    }
    
    private int pedSpawner_timer = PED_TIMER_BASE;
    private void pedSpawner(){
	float spawnrate = renderer.getPeds();
	if( spawnrate < 1.0f )
	    return;
        
        pedSpawner_timer--;
	if( pedSpawner_timer > 0 )
	    return;
	pedSpawner_timer = PED_TIMER_BASE + (int)(PED_TIMER * (100.0f - spawnrate) ) / 100;
	
	int rand;
	
	// Starting direction
	rand = (int)(rand_uniform.nextFloat() * 4);
	int dir_start;
	switch (rand) {
	    case 0:
		dir_start = Dir.UP;
		break;
	    case 1:
		dir_start = Dir.DOWN;
		break;
	    case 2:
		dir_start = Dir.RIGHT;
		break;
	    default:
		dir_start = Dir.LEFT;
		break;
	}
	
	// Relative direction
	rand = (int)(rand_uniform.nextFloat() * 3);
	int dir_rel;
	switch (rand) {
	    case 0:
		dir_rel = Dir.GO_LEFT;
		break;
	    case 1:
		dir_rel = Dir.GO_RIGHT;
		break;
	    default:
		dir_rel = Dir.GO_FORWARD;
		break;
	}
	
	peds.add(new Ped(lights, PED_SIZE,dir_start, dir_rel, renderer.loopCount));
    }
    
    private float nextNonUniformRandom(){
	float gaus =(float) rand_uniform.nextGaussian();
	// Scale down to get standard deviation 1/6
	gaus = gaus/6;
	// Move mean to 0.5
	gaus += 0.5;
	// Lower limit
	if(gaus < 0)
	    gaus = 0;
	/*
	* Removed upper limit to increase randomness of spawning
	    // Upper limit
	    if(gaus >= 1)
		gaus = 0.999f; // 1 is exclusive
	*/
        return gaus;
	
	// OLD APPROCHIMATION:
	// float f = (float) Math.random();
	/*
	*   a*x^3
	*   a*(x-0.5)^3 + 0.5 // center graph
	*   x=0,y=0 is a point
	*   a*(-0.5)^3 + 0.5 = 0
	*   a*0.5^3 = 0.5
	*   a = 4
	*
	*   4*(x-0.5)^3 + 0.5
	*/
	//return 4.0f*(f-0.5f)*(f-0.5f)*(f-0.5f)+0.5f;
    }
    
    public static float dis(float x0,float y0,float x1,float y1){
	float dx = x0-x1;
	float dy = y0-y1;
	float sq = dx*dx+dy*dy;
	return (float)Math.sqrt((double)sq);
    }
    public static float disSq(float x0,float y0,float x1,float y1){
	float dx = x0-x1;
	float dy = y0-y1;
	float sq = dx*dx+dy*dy;
	return sq;
    }
}
